package com.xypower.mpmaster;

import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.logging.Logger;


public class SingletonThread {
    private static volatile SingletonThread instance;
    private final ExecutorService executor;
    private final AtomicBoolean isRunning = new AtomicBoolean(false);
    private final Object lock = new Object();

    private SingletonThread() {
        // 使用单线程池，复用线程
        executor = Executors.newSingleThreadExecutor(r -> {
            Thread thread = new Thread(r, "OptimizedSingletonThread");
            thread.setPriority(Thread.NORM_PRIORITY - 1); // 稍低优先级
            return thread;
        });
    }

    // 双重检查锁定单例模式
    public static SingletonThread getInstance() {
        if (instance == null) {
            synchronized (SingletonThread.class) {
                if (instance == null) {
                    instance = new SingletonThread();
                }
            }
        }
        return instance;
    }

    /**
     * 执行任务（线程安全且节省资源）
     *
     * @param task 要执行的任务
     * @return true表示任务已接受执行，false表示跳过执行
     */
    public boolean execute(Runnable task) {
        if (task == null) {
            throw new IllegalArgumentException("Task cannot be null");
        }

        synchronized (lock) {
            if (isRunning.get()) {
                return false;
            }
            isRunning.set(true);
        }

        executor.execute(() -> {
            try {
                task.run();
            } finally {
                synchronized (lock) {
                    isRunning.set(false);
                }
            }
        });

        return true;
    }

    /**
     * 检查是否有任务正在执行
     */
    public boolean isBusy() {
        synchronized (lock) {
            return isRunning.get();
        }
    }

    /**
     * 关闭线程池（在应用退出时调用）
     */
    public void shutdown() {
        executor.shutdown();
    }
}